package mixins

import (
	"context"
	"entgo.io/ent"
	"entgo.io/ent/dialect/sql"
	"entgo.io/ent/schema/field"
	"entgo.io/ent/schema/mixin"
	"fmt"
	"git.noahlan.cn/noahlan/ntool-biz/core/orm/nent/hook"
)

// OptimisticLockMixin implements the ent.Mixin for sharing
// OptimisticLock fields with package schemas.
type OptimisticLockMixin struct {
	mixin.Schema
}

func (OptimisticLockMixin) Fields() []ent.Field {
	return []ent.Field{
		field.Int64("version").
			Default(1).
			Comment("optimistic lock | 乐观锁"),
	}
}

type optimisticLockKey struct{}

// WithVersion returns a new context that contains pre-version value for interceptor/mutators.
func WithVersion(parent context.Context, version int64) context.Context {
	return context.WithValue(parent, optimisticLockKey{}, version)
}

// Hooks of the OptimisticLockMixin.
func (d OptimisticLockMixin) Hooks() []ent.Hook {
	return []ent.Hook{
		hook.On(func(next ent.Mutator) ent.Mutator {
			return ent.MutateFunc(func(ctx context.Context, m ent.Mutation) (ent.Value, error) {
				mx, ok := m.(interface {
					AddVersion(i int64)
					WhereP(...func(*sql.Selector))
				})
				if !ok {
					return nil, fmt.Errorf("unexpected mutation type %T", m)
				}

				// get pre-version for update condition | where version={pre-version}
				if preVersion, preVersionOk := ctx.Value(optimisticLockKey{}).(int64); preVersionOk {
					d.P(mx, preVersion)
				}
				// set version=version+1
				mx.AddVersion(1)
				return next.Mutate(ctx, m)
			})
		}, ent.OpUpdate|ent.OpUpdateOne),
	}
}

// P adds a storage-level predicate to the queries and mutations.
func (d OptimisticLockMixin) P(w interface{ WhereP(...func(*sql.Selector)) }, preVersion int64) {
	w.WhereP(
		sql.FieldEQ(d.Fields()[0].Descriptor().Name, preVersion),
	)
}
