package katago

import (
	"errors"
	"git.noahlan.cn/noahlan/ntool/ncmd"
	"git.noahlan.cn/noahlan/ntool/nlog"
	"git.noahlan.cn/noahlan/ntool/nstr"
	"git.noahlan.cn/noahlan/ntool/nsys/atomic"
	"sync"
	"time"
)

var (
	DefaultMaxMessageId uint64 = 9999999
	DefaultTimeout             = 10 * time.Minute
)

var ErrTimeout = errors.New("timeout")

type (
	Options struct {
		DevMode      bool          `json:",default=false"`   // 开发模式
		MaxMessageId uint64        `json:",default=9999999"` // 最大消息ID
		Timeout      time.Duration `json:""`                 // 单条命令最大等待时间，默认10min
	}
	Option func(options *Options)

	pendingMsg struct {
		chWait   chan struct{}
		respJson string
	}

	AnalysisEngine struct {
		*Options

		Cmd        *ncmd.Cmd
		serializer *Serializer
		mid        *atomic.AtomicInt64

		pendingMsg map[string]*pendingMsg

		mu sync.RWMutex
	}
)

func NewKatagoAnalysisEngine(opts ...Option) *AnalysisEngine {
	serializer := NewSerializer()
	ret := &AnalysisEngine{
		Options: &Options{
			DevMode: false,
		},
		mid:        atomic.NewAtomicInt64(),
		pendingMsg: make(map[string]*pendingMsg),

		mu: sync.RWMutex{},
	}
	for _, opt := range opts {
		opt(ret.Options)
	}
	if ret.MaxMessageId == 0 {
		ret.MaxMessageId = DefaultMaxMessageId
	}
	if ret.Timeout == 0 {
		ret.Timeout = DefaultTimeout
	}

	ret.serializer = serializer
	ret.Cmd = ncmd.NewCmd(ncmd.WithOptions(&ncmd.Options{
		Marshaler:      serializer,
		Buffered:       false,
		CombinedOutput: false,
		Streaming:      true,
		LineBufferSize: ncmd.DEFAULT_LINE_BUFFER_SIZE,
		DevMode:        ret.DevMode,
	}))

	return ret
}

func (e *AnalysisEngine) Start(name string, args ...string) <-chan ncmd.Status {
	statusChan := e.Cmd.Start(name, args...)

	go e.handleMessage()

	return statusChan
}

func (e *AnalysisEngine) StartSafe(name string, args ...string) <-chan ncmd.Status {
	ret := e.Cmd.StatusChan()
	if !e.Cmd.Started() {
		if !e.Cmd.Stopped() {
			ret = e.Start(name, args...)
		} else {
			// 被停止
			e.Cmd = e.Cmd.Clone()
			ret = e.Start(name, args...)
		}
	}

	return ret
}

func (e *AnalysisEngine) Send(data BaseMsg, resp any) error {
	pMsg := &pendingMsg{
		chWait: make(chan struct{}, 1),
	}
	id := e.nextId()
	data.SetMessageId(id)

	e.mu.Lock()
	e.pendingMsg[id] = pMsg
	e.mu.Unlock()

	err := e.Cmd.Send(data)
	if err != nil {
		return err
	}

	// waiting pending message with timeout
	timer := time.NewTimer(e.Timeout)

	select {
	case <-timer.C:
		// error
		e.mu.Lock()
		delete(e.pendingMsg, id)
		e.mu.Unlock()

		return ErrTimeout
	case <-pMsg.chWait:
		ok, err := e.serializer.Unmarshal(pMsg.respJson, resp)
		if err != nil {
			return err
		}
		if !ok {
			return errors.New("消息不完整")
		}
		return nil
	}
}

func (e *AnalysisEngine) nextId() string {
	idInt := e.mid.IncrementAndGet()
	if idInt > int64(e.MaxMessageId) {
		e.mid.Reset()
		idInt = e.mid.IncrementAndGet()
	}
	return nstr.SafeString(idInt)
}

func (e *AnalysisEngine) handleMessage() {
	for e.Cmd.Stdout != nil || e.Cmd.Stderr != nil {
		select {
		case line, open := <-e.Cmd.Stdout:
			if !open {
				e.Cmd.Stdout = nil
				continue
			}
			//if e.DevMode {
			//	nlog.Debugf("接收单条消息: %s", line)
			//}
			var baseResp BaseResp
			ok, err := e.serializer.Unmarshal(line, &baseResp)
			if err != nil {
				nlog.Errorf("解释json失败 %+v", err)
				continue
			}
			if e.DevMode {
				nlog.Debugf("接收完整消息: %+v", e.serializer.contentBuf.String())
			}

			e.mu.RLock()
			pMsg, ok := e.pendingMsg[baseResp.ID]
			e.mu.RUnlock()
			if !ok {
				continue
			}

			pMsg.respJson = e.serializer.LastJson()
			pMsg.chWait <- struct{}{}
		case line, open := <-e.Cmd.Stderr:
			if !open {
				e.Cmd.Stdout = nil
				continue
			}
			nlog.Errorf("错误消息: %s\n", line)
		}
	}
}

func (e *AnalysisEngine) check(err error, req any) bool {
	if err != nil {
		nlog.Errorf("发送命令[%+v]失败 %v", req, err)
		return false
	}
	return true
}

////////////////////////////////

// QueryVersion 查询版本号
func (e *AnalysisEngine) QueryVersion(req *QueryActionReq) (*QueryVersionResp, error) {
	var resp QueryVersionResp

	err := e.Send(req, &resp)
	if !e.check(err, req) {
		return nil, err
	}
	return &resp, nil
}

// Query 分析查询
func (e *AnalysisEngine) Query(req *QueryReq) (*QueryResp, error) {
	var resp QueryResp

	err := e.Send(req, &resp)
	if !e.check(err, req) {
		return nil, err
	}
	return &resp, nil
}

// ClearCache 清除缓存
func (e *AnalysisEngine) ClearCache() (*ClearCacheResp, error) {
	var resp ClearCacheResp
	req := &QueryActionReq{
		Action: AnalActionClearCache,
	}
	err := e.Send(req, &resp)
	if !e.check(err, req) {
		return nil, err
	}
	return &resp, nil
}

// Terminate 终止某次分析
// terminateId: 分析轮次ID
// turnNumbers: 需要终止分析的回合数（手数），可空：所有回合
func (e *AnalysisEngine) Terminate(terminateId string, turnNumbers []int32) (*TerminateResp, error) {
	var resp TerminateResp
	req := &QueryActionReq{
		Action:      AnalActionTerminate,
		TerminateId: terminateId,
		TurnNumbers: turnNumbers,
	}
	err := e.Send(req, &resp)
	if !e.check(err, req) {
		return nil, err
	}
	return &resp, nil
}

// TerminateAll 终止所有分析
// turnNumbers: 需要终止所有分析轮次的回合数，可空：所有回合
func (e *AnalysisEngine) TerminateAll(turnNumbers []int32) (*TerminateAllResp, error) {
	var resp TerminateAllResp
	req := &QueryActionReq{
		Action:      AnalActionTerminateAll,
		TurnNumbers: turnNumbers,
	}
	err := e.Send(req, &resp)
	if !e.check(err, req) {
		return nil, err
	}
	return &resp, nil
}
